"""
Implementation of Moran spectral randomization.
"""

# Author: Oualid Benkarim <oualid.benkarim@mcgill.ca>
# License: BSD 3 clause


import numpy as np
import scipy.sparse as ssp
from scipy.spatial.distance import cdist

from sklearn.utils import check_random_state
from sklearn.base import BaseEstimator



def compute_mem(w, n_ring=1, spectrum='nonzero', tol=1e-10):
    """ Compute Moran eigenvectors map.

    Parameters
    ----------
    w : BSPolyData, ndarray or sparse matrix, shape = (n_vertices, n_vertices)
        Spatial weight matrix or surface. If surface, the weight matrix is
        built based on the inverse geodesic distance between each vertex
        and the vertices in its `n_ring`.
    n_ring : int, optional
        Neighborhood size to build the weight matrix. Only used if user
        provides a surface mesh. Default is 1.
    spectrum : {'all', 'nonzero'}, optional
        Eigenvalues/vectors to select. If 'all', recover all eigenvectors
        except the smallest one. Otherwise, select all except non-zero
        eigenvectors. Default is 'nonzero'.
    tol : float, optional
        Minimum value for an eigenvalue to be considered non-zero.
        Default is 1e-10.

    Returns
    -------
    w : 1D ndarray, shape (n_components,)
        Eigenvalues in descending order. With ``n_components = n_vertices - 1``
        if ``spectrum == 'all'`` and ``n_components = n_vertices - n_zero`` if
        ``spectrum == 'nonzero'``, and `n_zero` is number of zero eigenvalues.
    mem : 2D ndarray, shape (n_vertices, n_components)
        Eigenvectors of the weight matrix in same order.

    See Also
    --------
    :func:`.moran_randomization`
    :class:`.MoranRandomization`

    References
    ----------
    * Wagner H.H. and Dray S. (2015). Generating spatially constrained
      null models for irregularly spaced data using Moran spectral
      randomization methods. Methods in Ecology and Evolution, 6(10):1169-78.

    """

    if spectrum not in ['all', 'nonzero']:
        raise ValueError("Unknown autocor '{0}'.".format(spectrum))


    # Doubly centering weight matrix
    if ssp.issparse(w):
        m = w.mean(axis=0).A
        wc = w.mean() - m - m.T

        if not ssp.isspmatrix_coo(w):
            w_format = w.format
            w = w.tocoo(copy=False)
            row, col = w.row, w.col
            w = getattr(w, 'to' + w_format)(copy=False)
        else:
            row, col = w.row, w.col
        wc[row, col] += w.data

    else:
        m = w.mean(axis=0, keepdims=True)
        wc = w.mean() - m - m.T
        wc += w

    # when using float64, eigh is unstable for sparse matrices
    ev, mem = np.linalg.eigh(wc.astype(np.float32))
    ev, mem = ev[::-1], mem[:, ::-1]

    # Remove zero eigen-value/vector
    ev_abs = np.abs(ev)
    mask_zero = ev_abs < tol
    n_zero = np.count_nonzero(mask_zero)

    # Multiple zero eigenvalues
    if spectrum == 'all':
        if n_zero > 1:
            n = w.shape[0]
            memz = np.hstack([mem[:, mask_zero], np.ones((n, 1))])
            q, _ = np.linalg.qr(memz)
            mem[:, mask_zero] = q[:, :-1]
            idx_zero = mask_zero.argmax()
        else:
            idx_zero = ev_abs.argmin()

        ev[idx_zero:-1] = ev[idx_zero+1:]
        mem[:, idx_zero:-1] = mem[:, idx_zero + 1:]
        ev = ev[:-1]
        mem = mem[:, :-1]

    else:  # only nonzero
        mask_nonzero = ~mask_zero
        ev = ev[mask_nonzero]
        mem = mem[:, mask_nonzero]

    return mem, ev


def moran_randomization(x, mem, n_rep=100, procedure='singleton', joint=False,
                        random_state=None):
    """ Generate random samples from `x` based on Moran spectral randomization.

    Parameters
    ----------
    x : 1D or 2D ndarray, shape = (n_vertices,) or (n_vertices, n_feat)
        Array of variables arranged in columns, where `n_feat` is the number
        of variables.
    mem : 2D ndarray, shape = (n_vertices, nv)
        Moran eigenvectors map, where `nv` is the number of eigenvectors
        arranged in columns.
    n_rep : int, optional
        Number of random samples. Default is 100.
    procedure : {'singleton, 'pair'}, optional
        Procedure to generate the random samples. Default is 'singleton'.
    joint : boolean, optional
        If True variables are randomized jointly. Otherwise, each variable is
        randomized separately. Default is False.
    random_state : int or None, optional
        Random state. Default is None.

    Returns
    -------
    output : ndarray, shape = (n_rep, n_vertices, n_feat)
        Random samples. If ``n_feat == 1``, shape = (n_rep, n_vertices).

    See Also
    --------
    :func:`.compute_mem`
    :class:`.MoranRandomization`

    References
    ----------
    * Wagner H.H. and Dray S. (2015). Generating spatially constrained
      null models for irregularly spaced data using Moran spectral
      randomization methods. Methods in Ecology and Evolution, 6(10):1169-78.

    """

    if x.ndim == 1:
        x = np.atleast_2d(x).T

    procedure = procedure.lower()
    if procedure not in ['singleton', 'pair']:
        raise ValueError("Unknown procedure '{0}'".format(procedure))

    rs = check_random_state(random_state)

    n_comp = mem.shape[1]
    n_rows = x.shape[0]
    n_cols = 1 if joint else x.shape[1]

    rxv = 1 - cdist(x.T, mem.T, 'correlation').T
    if procedure == 'singleton':
        rxv2 = rxv * rs.choice([-1., 1.], size=(n_rep, n_comp, n_cols))

    else:  # pair
        n_pairs = n_comp // 2
        n_top = 2 * n_pairs
        is_odd = n_top != n_comp

        rsq = rxv ** 2
        rxv2 = np.empty((n_rep,) + rxv.shape)
        for i in range(n_rep):
            p = rs.permutation(n_comp)
            # ia, ib = p[:n_top:2], p[1:n_top:2]
            ia, ib = p[:n_pairs], p[n_pairs:n_top]

            if is_odd:  # singleton method for last item
                rxv2[i, p[-1]] = rxv[p[-1]] * rs.choice([-1, 1], size=n_cols)

            phi = rs.uniform(0, 2 * np.pi, size=(n_pairs, n_cols))
            if joint:
                phi = phi + np.arctan2(rxv[ia], rxv[ib])
            rxv2[i, ia] = rxv2[i, ib] = np.sqrt(rsq[ia] + rsq[ib])
            rxv2[i, ia] *= np.cos(phi)
            rxv2[i, ib] *= np.sin(phi)

    x_mean = x.mean(axis=0)
    x_std = x.std(axis=0, ddof=1)
    sim = x_mean + (mem @ rxv2) * (np.sqrt(n_rows - 1) * x_std)

    return sim.squeeze()


class MoranRandomization(BaseEstimator):
    """ Moran spectral randomization.

    Parameters
    ----------
    procedure : {'singleton, 'pair'}, optional
        Procedure to generate the random samples. Default is 'singleton'.
    spectrum : {'all', 'nonzero'}, optional
        Eigenvalues/vectors to select. If 'all', recover all eigenvectors
        except one. Otherwise, select all except non-zero eigenvectors.
        Default is 'nonzero'.
    joint : boolean, optional
        If True variables are randomized jointly. Otherwise, each variable is
        randomized separately. Default is False.
    n_rep : int, optional
        Number of randomizations. Default is 100.
    n_ring : int, optional
        Neighborhood size to build the weight matrix. Only used if user provides
        a surface mesh. Default is 1.
    tol : float, optional
        Minimum value for an eigenvalue to be considered non-zero.
        Default is 1e-10.
    random_state : int or None, optional
        Random state. Default is None.

    Attributes
    ----------
    mev_ : 1D ndarray, shape (n_components,)
        Eigenvalues of the weight matrix in descending order.
    mem_ : 2D ndarray, shape (n_vertices, n_components)
        Eigenvectors of the weight matrix in same order.

    See Also
    --------
    :class:`.SpinPermutations`

    """

    def __init__(self, procedure='singleton', spectrum='nonzero', joint=False,
                 n_rep=100, n_ring=1, tol=1e-10, random_state=None):

        self.procedure = procedure
        self.spectrum = spectrum
        self.joint = joint
        self.n_rep = n_rep
        self.n_ring = n_ring
        self.tol = tol
        self.random_state = random_state

    def fit(self, w):
        """ Compute Moran eigenvectors map.

        Parameters
        ----------
        w : BSPolyData, ndarray or sparse matrix, shape = (n_verts, n_verts)
            Spatial weight matrix or surface. If surface, the weight matrix is
            built based on the inverse geodesic distance between each vertex
            and the vertices in its `n_ring`.

        Returns
        -------
        self : object
            Returns self.

        """

        self.mem_, self.mev_ = compute_mem(w, spectrum=self.spectrum,
                                           tol=self.tol)
        return self

    def randomize(self, x):
        """ Generate random samples from `x`.

        Parameters
        ----------
        x : 1D or 2D ndarray, shape = (n_verts,) or (n_verts, n_feat)
            Array of variables arranged in columns, where `n_feat` is the
            number of variables.

        Returns
        -------
        output : ndarray, shape = (n_rep, n_verts, n_feat)
            Random samples. If ``n_feat == 1``, shape = (n_rep, n_verts).

        """

        rand = moran_randomization(x, self.mem_, n_rep=self.n_rep,
                                   procedure=self.procedure, joint=self.joint,
                                   random_state=self.random_state)
        return rand



import numpy as np
import nibabel as nib
import pandas as pd
from tabulate import tabulate
from scipy.ndimage import rotate, shift
from scipy.stats import spearmanr, pearsonr
from joblib import Parallel, delayed
from sklearn.metrics import adjusted_mutual_info_score, adjusted_rand_score
import warnings
import hippomaps.utils
import hippomaps.config
import matplotlib.pyplot as plt
from brainspace.mesh.mesh_io import read_surface
from hippomaps.moran import MoranRandomization
from brainspace.mesh import mesh_elements as me
from eigenstrapping import SurfaceEigenstrapping
from pathlib import Path
resourcesdir=str(Path(__file__).parents[1]) + '/resources'


def spin_test(imgfix, imgperm, nperm=1000, metric='pearsonr', label='hipp', den='0p5mm'):
    """
       Permutation testing of unfolded hippocampus maps.

       Original code by Bradley Karat at https://github.com/Bradley-Karat/Hippo_Spin_Testing
       Karat, B. G., DeKraker, J., Hussain, U., Köhler, S., & Khan, A. R. (2023).
       Mapping the macrostructure and microstructure of the in vivo human hippocampus using diffusion MRI.
       Human Brain Mapping. https://doi.org/10.1002/hbm.26461

       Parameters
       ----------
       imgfix : str or array
           Path to the fixed map, or fixed array map.
       imgperm : str or array
           Path to the fixed map to be permuted, or array map to be permuted.
       nperm : int
           Number of permutations to perform.
       metric : str, optional
           Metric for comparing maps (one of pearsonr, spearmanr, adjusted_mutual_info_score, or adjusted_mutual_info_score).
           Default is 'pearson'.
       label : str, optional
           Label for the hippocampus ('hipp' or 'dentate'). Default is 'hipp'.
       den : str, optional
           Density of the surface data. Default '0p5mm'.

       Returns
       -------
       metricnull : Null distribution of the specified metric
       permutedimg : All permuted spatial maps at 'unfoldiso' density.
       r_obs :  The observed association between the two aligned maps.
       pval : p-value based on metricnull r_obs.
       """
    if type(imgfix) == str:
        imgfix = nib.load(imgfix).agg_data()
    if type(imgperm) == str:
        imgperm = nib.load(imgperm).agg_data()

    # resmaple to space-unfoldiso
    if den != 'unfoldiso':
        imgperm = hippomaps.utils.density_interp(den, 'unfoldiso', imgperm, label=label, method='nearest')[0]
        imgfix = hippomaps.utils.density_interp(den, 'unfoldiso', imgfix, label=label, method='nearest')[0]  
    if label == 'hipp':
        imgperm = np.reshape(imgperm, (126, 254))  # get maps to 126x254
        imgfix = np.reshape(imgfix, (126, 254))  # get maps to 126x254
    elif label == 'dentate':
        imgperm = np.reshape(imgperm, (32, 254)) # get maps to 32x254
        imgfix = np.reshape(imgfix, (32, 254))

    rotation = np.random.randint(1, 360, nperm)  # generate random rotations
    translate1 = np.random.randint(-63, 64, nperm)  # generate random translations
    translate2 = np.random.randint(-127, 128, nperm)
    permutedimg = np.empty((126, 254, nperm))
    metricnull = np.empty((nperm))

    for ii in range(nperm):
        rotimg = rotate(imgperm, rotation[ii], axes=(1, 0), reshape=False, output=None, order=3, mode='wrap',
                        cval=0.0, prefilter=True)  # rotate image
        transrotimg = shift(rotimg, [translate1[ii], translate2[ii]], output=None, order=3, mode='wrap', cval=0.0,
                            prefilter=True)  # translate image
        permutedimg[:, :, ii] = transrotimg  # this is our permuted image at unfoldiso density

    r_obs = pearsonr(imgfix.flatten(), imgperm.flatten())[0]  
    for ii in range(nperm):
        metricnull[ii] = eval(metric)(imgfix.flatten(), permutedimg[:,:,ii].flatten())[0]  
    
    # p-value is the sum of all instances where null correspondance is >= observed correspondance / nperm
    pval = np.mean(np.abs(metricnull) >= np.abs(r_obs))  

    return metricnull, permutedimg, pval, r_obs


def moran_test(imgfix, imgperm, nperm=1000, metric='pearsonr', label='hipp', den='0p5mm'):
    """
       Moran Spectral Randomization
       Moran Spectral Randomization (MSR) computes Moran’s I, a metric for spatial auto-correlation and generates normally distributed data with similar auto-correlation. MSR relies on a weight matrix denoting the spatial proximity of features to one another. Within neuroimaging, one straightforward example of this is inverse geodesic distance i.e. distance along the cortical surface.

       Code from BrainSpace https://brainspace.readthedocs.io/en/latest/python_doc/auto_examples/plot_tutorial3.html
       Vos de Wael, Reinder, Oualid Benkarim, Casey Paquola, Sara Lariviere, Jessica Royer, Shahin Tavakol, Ting Xu et al. "BrainSpace: a toolbox for the analysis of macroscale gradients in neuroimaging and connectomics datasets." Communications biology 3, no. 1 (2020): 103.

       Parameters
       ----------
       imgfix : str or array
           Path to the fixed map, or fixed array map.
       imgperm : str or array
           Path to the fixed map to be permuted, or array map to be permuted.
       nperm : int
           Number of permutations to perform.
       metric : str, optional
           Metric for comparing maps (one of pearsonr, spearmanr).
           Default is 'pearsonr'.
       label : str, optional
           Label for the hippocampus ('hipp' or 'dentate'). Default is 'hipp'.
       den : str, optional
           Density of the surface data. Default '0p5mm'.

       Returns
       -------
       metricnull : Null distribution of the specified metric
       permutedimg : All permuted spatial maps at 'unfoldiso' density.
       r_obs :  The observed association between the two aligned maps.
       pval : p-value based on metricnull r_obs.
       """
    if type(imgfix) == str:
        imgfix = nib.load(imgfix).agg_data().flatten()
    if type(imgperm) == str:
        imgperm = nib.load(imgperm).agg_data().flatten()
    
    # load reference surface to get geodesic distance
    surf = read_surface(f"{resourcesdir}/canonical_surfs/tpl-avg_space-canonical_den-{den}_label-{label}_midthickness.surf.gii")
    # wrap convenient brainspace function for weights (geodesic distance) and MRS
    weights = me.get_ring_distance(surf, n_ring=1)
    weights.data **= -1
    msr = MoranRandomization(n_rep=nperm, spectrum='all')
    msr.fit(weights)

    # get observed correlation
    r_obs = eval(metric)(imgfix, imgperm)[0]

    # randomize
    imgperm_rand = msr.randomize(imgperm)
    metricnull = np.ones((nperm))*np.nan
    for d in range(nperm):
        try:
            metricnull[d] = eval(metric)(imgfix, imgperm_rand[d,:])[0]
        except: 
            warnings.warn(f"permuation {d} contains a NaN or Inf")

    # p-value is the sum of all instances where null correspondance is >= observed correspondance / nperm
    pval = np.nanmean(np.abs(metricnull) >= np.abs(r_obs))  

    return metricnull, imgperm_rand, pval, r_obs

def eigenstrapping(imgfix, imgperm, nperm=1000, metric='pearsonr', label='hipp', den='0p5mm', num_modes=200, permute=False, resample=False, **qwargs):
    """
        Awesome new tool at https://www.biorxiv.org/content/10.1101/2024.02.07.579070v1.abstract
        Generates null models of spatial maps by rotating geometric eigenmodes.
        Koussis, N. C., Pang, J. C., Jeganathan, J., Paton, B., Fornito, A., Robinson, P. A., ... & Breakspear, M. (2024). Generation of surrogate brain maps preserving spatial autocorrelation through random rotation of geometric eigenmodes. bioRxiv, 2024-02.
       
        Parameters
        ----------
        imgfix : str or array
           Path to the fixed map, or fixed array map.
        imgperm : str or array
           Path to the fixed map to be permuted, or array map to be permuted.
        nperm : int
            Number of permutations to perform.
        metric : str, optional
            Metric for comparing maps (one of pearsonr, spearmanr).
            Default is 'pearsonr'.
        label : str, optional
            Label for the hippocampus ('hipp' or 'dentate'). Default is 'hipp'.
        den : str, optional
            Density of the surface data. Default '0p5mm'.
        num_modes : int, optional
            Number of eigenmodes to use. Default is 200.
        permute : bool, optional
            Set whether to permute surrogate map from original map to preserve values. Default is False.
        resample : bool, optional
            Set whether to resample surrogate map from original map to preserve values. Default is False.
        **qwargs : dict, optional
            Additional keyword arguments for customization.
            See https://eigenstrapping.readthedocs.io/en/latest/generated/eigenstrapping.SurfaceEigenstrapping.html#eigenstrapping.SurfaceEigenstrapping

        Returns
        -------
        metricnull : Null distribution of the specified metric
        permutedimg : All permuted spatial maps at 'unfoldiso' density.
        r_obs :  The observed association between the two aligned maps.
        pval : p-value based on metricnull r_obs.
        """
    if type(imgfix) == str:
        imgfix = nib.load(imgfix).agg_data().flatten()
    if type(imgperm) == str:
        imgperm = nib.load(imgperm).agg_data().flatten()
    
    # load reference surface and put in Eigenstrapping class
    eigen = SurfaceEigenstrapping(surface=f"{resourcesdir}/canonical_surfs/tpl-avg_space-canonical_den-{den}_label-{label}_midthickness.surf.gii", 
        data=imgperm, num_modes=num_modes, permute=False, resample=False, **qwargs)

    # get observed correlation
    r_obs = eval(metric)(imgfix, imgperm)[0]

    # randomize
    hippomaps.utils.blockPrint()
    imgperm_rand = eigen(n=nperm)
    hippomaps.utils.enablePrint()
    metricnull = np.ones((nperm))*np.nan
    for d in range(nperm):
        try:
            metricnull[d] = eval(metric)(imgfix, imgperm_rand[d,:])[0]
        except: 
            warnings.warn(f"permuation {d} contains a NaN or Inf")

    # p-value is the sum of all instances where null correspondance is >= observed correspondance / nperm
    pval = np.nanmean(np.abs(metricnull) >= np.abs(r_obs))  

    return metricnull, imgperm_rand, pval, r_obs

def contextualize2D(taskMaps, n_topComparison=3, nperm=1000, plotTable=True, plot2D=True):
    """
       Compares the present maps to thoes in the inital HippoMaps release

       Parameters
       ----------
       taskMaps : a VxT matrix of intensities. T is the number of maps to examine
       n_topComparison : int, optional
           Top N features for which to reurn comparisons (Pearson's R and p)
       plot2D : bool
           Whether to plot the new maps in context of other maps

       Returns
       -------
       topFeatures : Names of most similar intialized features
       topR : Correpsonding Pearson's R values
       topP : Corresponding p values (spin test)
       APcorr : Correlation with AP coordinates
       Subfscorr : Max correlation with subfields (Spearmann)
       ax : axis handle
       """
    if taskMaps.ndim==1: taskMaps = taskMaps.reshape([-1,1])
    nT = taskMaps.shape[1]

    # load required data
    contextHM = np.load(f'{resourcesdir}/2Dcontextualize/initialHippoMaps.npz')
    
    # resample all input data to 0p5mm (if needed)
    nV,iV = hippomaps.config.get_nVertices(['hipp'],'0p5mm')
    if taskMaps.shape[0] != nV:
        taskMapsresamp,_,_ = hippomaps.utils.density_interp(hm.config.get_label_from_nV(taskMaps.shape[0])[1],'0p5mm',taskMaps, label='hipp')
    else:
        taskMapsresamp = taskMaps

    # permute the input task maps (this is generally faster than permuting the reference features!)
    permutedTasks = np.zeros((nV,nT,nperm))
    metricnull = np.ones((nT,nperm))*np.nan
    hippomaps.utils.blockPrint()
    for t in range(nT):
            eigen = SurfaceEigenstrapping(surface=f"{resourcesdir}/canonical_surfs/tpl-avg_space-canonical_den-0p5mm_label-hipp_midthickness.surf.gii", 
                data=taskMapsresamp[:,t])
            permutedTasks[:,t,:] = eigen(nperm).T
    hippomaps.utils.enablePrint()
    # approx 9m to here

    # compute correlations with AP and all subfield permutations
    APR = np.ones((nT))*np.nan
    APp = np.ones((nT))*np.nan
    Subsp = np.ones((nT))*np.nan
    SubsR = np.ones((nT))*np.nan
    for t in range(nT):
        APR[t] = pearsonr(taskMapsresamp[:,t],contextHM['AP'])[0]
        SubsWithTask = pd.DataFrame(np.concatenate((taskMapsresamp[:,t].reshape([nV,1]),contextHM['subfields_permuted']),axis=1))
        Subfscorr = SubsWithTask.corr('spearman').to_numpy()[0,1:]
        Subfscorr_best = np.nanargmax(np.abs(Subfscorr))
        SubsR[t] = Subfscorr[Subfscorr_best]
        metricnull = np.zeros((nperm))
        for d in range(nperm):
            metricnull[d] = pearsonr(contextHM['AP'], permutedTasks[:,t,d])[0]
        APp[t] = np.nanmean(np.abs(metricnull) >= np.abs(APR[t]))
        metricnull = np.zeros((nperm))
        for d in range(nperm):
            metricnull[d] = spearmanr(contextHM['subfields_permuted'][:,Subfscorr_best], permutedTasks[:,t,d])[0]
        Subsp[t] = np.nanmean(np.abs(metricnull) >= np.abs(SubsR[t]))
    # 18m to here

    # compute correlations with extant features (but only the top 3 to save time)
    topFeatures = np.ones((nT,n_topComparison), dtype='object')
    topR = np.ones((nT,n_topComparison))*np.nan
    topP = np.ones((nT,n_topComparison))*np.nan
    if n_topComparison >0:
        p = np.zeros((nT,len(contextHM['features'])))
        R = np.zeros((nT,len(contextHM['features'])))
        for t in range(nT):
            for j in range(len(contextHM['features'])):
                R[t,j],p[t,j] = pearsonr(taskMapsresamp[:,t],contextHM['featureData'][:,j])
        # get ordering of the closest n_topComparison neighbours and then run spin test
        for t in range(nT):
            order = np.argsort(np.abs(R[t,:]))[::-1]
            for c in range(n_topComparison):
                topFeatures[t,c] = contextHM['features'][order[c]]
                topR[t,c] = R[t,order[c]]
                metricnull = np.zeros((nperm))
                for d in range(nperm):
                    metricnull[d] = pearsonr(contextHM['featureData'][:,order[c]], permutedTasks[:,t,d])[0]
                topP[t,c] = np.nanmean(np.abs(metricnull) >= np.abs(topR[t,c]))

    # generate dict of results
    context2D = {"Input map": list(range(nT)), 
        "Subfields (R)": SubsR, 
        "Subfields (p corrected)": Subsp, 
        "A-P gradinet (R)": APR,
        "A-P gradinet (p corrected)": APp}
    for c in range(n_topComparison):
        context2D[str(c) +"th most correlated"] = topFeatures[:,c]
        context2D[str(c) +"th (R)"] = topR[:,c]
        context2D[str(c) +"th (p corrected)"] = topP[:,c]
    
    if plotTable:
        print(tabulate(context2D,headers="keys"))

    fig, ax = plt.subplots(figsize=(8,8))
    if plot2D:
        ax.spines[['right', 'top']].set_visible(False)
        ax.scatter(contextHM['axiscorrAPPD'][0],contextHM['subfieldsmaxcorr'],c=contextHM['colors'],cmap='Set3',s=200)
        plt.ylabel("absolute subfield correlation (Spearmann's R)")
        plt.xlabel("absolute AP correlation (Pearson's R)")
        for f,feature in enumerate(contextHM['features']):
            ax.annotate(str(int(contextHM['feature_n'][f])), (contextHM['axiscorrAPPD'][0,f]-.008, contextHM['subfieldsmaxcorr'][f]-.007))
        ax.scatter(np.abs(APR),np.abs(SubsR),color='k',s=200);
        for t in range(nT):
            ax.annotate(str(t), (np.abs(APR[t])-.008, np.abs(SubsR[t])-.007),color='w')
            
    return context2D, ax



import numpy as np
import nibabel as nib
from pathlib import Path
resourcesdir = str(Path(__file__).parents[1]) + '/resources'

labelNames = ['hipp','dentate']
densityNames = ['2mm','1mm','0p5mm','unfoldiso']

nVertices = np.zeros((len(labelNames), len(densityNames))).astype(int)
for l,label in enumerate(labelNames):
    for d,den in enumerate(densityNames):
        surf = nib.load(f'{resourcesdir}/canonical_surfs/tpl-avg_space-unfold_den-{den}_label-{label}_midthickness.surf.gii')
        nVertices[l,d] = int(surf.get_arrays_from_intent('NIFTI_INTENT_POINTSET')[0].data.shape[0])

def get_nVertices(labels,den):
    iLabels = np.where(np.in1d(labelNames, labels))[0]
    idens = np.where(np.in1d(densityNames, den))[0]
    nV = np.sum(nVertices[iLabels,idens])
    iV = []
    iV.append(range(nVertices[iLabels[0],idens[0]]))
    if len(iLabels)>1:
       iV.append(range(nVertices[iLabels[0],idens[0]], nV))
    return nV,iV

def get_label_from_nV(nV):
    lookupDen = dict()
    lookupLabel = dict()
    for l,label in enumerate(labelNames):
        for d,den in enumerate(densityNames):
            lookupDen[str(nVertices[l,d])] = densityNames[d]
            lookupLabel[str(nVertices[l,d])] = labelNames[l]
    return lookupLabel[str(nV)], lookupDen[str(nV)]


"""Neuroimaging statistics toolbox."""
__version__ = "0.1.3"

import hippomaps.config
import hippomaps.utils
import hippomaps.plotting
import hippomaps.stats
import hippomaps.fetcher

__all__ = ['config', 'utils', 'plotting', 'stats', 'fetcher']




import numpy as np
import nibabel as nib
import copy
import glob
from scipy.ndimage.filters import gaussian_filter
from numpy.matlib import repmat
from brainspace.mesh.mesh_io import read_surface
from brainspace.plotting import plot_hemispheres, plot_surf, build_plotter
from brainspace.mesh import mesh_creation as mc
from pathlib import Path

resourcesdir = str(Path(__file__).parents[1]) + '/resources'
import hippomaps.utils


def surfplot_canonical_foldunfold(cdata, hemis=['L', 'R'], labels=['hipp', 'dentate'], unfoldAPrescale=False,
                                  den='0p5mm', tighten_cwindow=False, resourcesdir=resourcesdir, size=[350, 300],
                                  **qwargs):
    """
       Plots canonical folded and unfolded surfaces for hippocampus and dentate gyrus.
       Parameters
       ----------
       cdata : numpy.ndarray
           Array with the shape Vx2xF, where V is the number of vertices (including DG unless specified),
           2 is the number of hemispheres (unless specified), and F is the number of rows/features.
       hemis : list of str, optional
           List of hemispheres to visualize. Default is ['L', 'R'].
       labels : list of str, optional
           List of labels for different structures. Default is ['hipp', 'dentate'].
       unfoldAPrescale : bool, optional
           Whether to pre-scale the anterior-posterior axis during unfolding. Default is False.
       den : str, optional
           Density parameter for surface plot. Default is '0p5mm'.
       tighten_cwindow : bool, optional
           Whether to tighten the color window for the surface plot. Default is False.
       resourcesdir : str, optional
           Directory path containing additional resources. Default is the value of resourcesdir.
       size : list of int, optional
           Size of the surface plot. Default is [350, 300].
       **qwargs : dict, optional
           Additional keyword arguments for customization.
           See https://brainspace.readthedocs.io/en/latest/generated/brainspace.plotting.surface_plotting.plot_surf.html#brainspace.plotting.surface_plotting.plot_surf

       Returns
       -------
       matplotlib.figure.Figure
           Figure object for the generated surface plot.
       Notes
       -----
       This function is suitable for plotting canonical folded and unfolded surfaces, and it is particularly useful
       when the data (`cdata`) isn't specific to one subject (e.g., maybe it has been averaged across many subjects).
       """

    # load surfaces
    rh = read_surface(
        f'{resourcesdir}/canonical_surfs/tpl-avg_space-canonical_den-{den}_label-hipp_midthickness.surf.gii')
    ru = read_surface(f'{resourcesdir}/canonical_surfs/tpl-avg_space-unfold_den-{den}_label-hipp_midthickness.surf.gii')
    ru.Points = ru.Points[:, [1, 0, 2]]  # reorient unfolded
    if unfoldAPrescale: ru.Points = area_rescale(ru.Points, den, 'hipp', APaxis=1)
    if len(labels) == 2:
        ud = read_surface(
            f'{resourcesdir}/canonical_surfs/tpl-avg_space-unfold_den-{den}_label-dentate_midthickness.surf.gii')
        hd = read_surface(
            f'{resourcesdir}/canonical_surfs/tpl-avg_space-canonical_den-{den}_label-dentate_midthickness.surf.gii')
        ud.Points = ud.Points[:, [1, 0, 2]]  # reorient unfolded
        ud.Points = ud.Points + [22, 0, 0]  # translate unfolded dg
        if unfoldAPrescale: ud.Points = area_rescale(ud.Points, den, 'dentate', APaxis=1)
        # add to original
        npts = rh.n_points
        rh = mc.build_polydata(np.concatenate((rh.Points.copy(), hd.Points.copy())),
                               cells=np.concatenate((rh.GetCells2D().copy(), hd.GetCells2D().copy() + npts)))
        ru = mc.build_polydata(np.concatenate((ru.Points.copy(), ud.Points.copy())),
                               cells=np.concatenate((ru.GetCells2D().copy(), ud.GetCells2D().copy() + npts)))

    # flip to get left hemisphere
    lh = mc.build_polydata(rh.Points.copy(), cells=rh.GetCells2D().copy())
    lh.Points[:, 0] = -lh.Points[:, 0]
    lu = mc.build_polydata(ru.Points.copy(), cells=ru.GetCells2D().copy())
    lu.Points[:, 0] = -lu.Points[:, 0]

    # do some cdata formatting
    cdata = np.reshape(cdata, [cdata.shape[0], len(hemis), -1])
    if len(cdata.shape) == 2: cdata = np.expand_dims(cdata, axis=2)
    if tighten_cwindow > 0:
        for i in range(0, cdata.shape[2]):
            cdata[:, :, i] = bound_cdata(cdata[:, :, i])

    # set up layout
    surfDict = {'Lf': lh, 'Lu': lu, 'Rf': rh, 'Ru': ru}
    surfList = np.ones((cdata.shape[2], len(hemis) * 2), dtype=object)
    arrName = np.ones((cdata.shape[2], len(hemis) * 2), dtype=object)
    for h, hemi in enumerate(hemis):
        if hemi == 'L':
            surfList[:, [h, h + 1]] = np.array([f"{hemi}f", f"{hemi}u"])
            for f in range(cdata.shape[2]):
                lh.append_array(cdata[:, h, f], name=f'feature{f}', at='point')
                lu.append_array(cdata[:, h, f], name=f'feature{f}', at='point')
        elif hemi == 'R':
            surfList[:, [h * 2, h * 2 + 1]] = np.array([f"{hemi}u", f"{hemi}f"])
            for f in range(cdata.shape[2]):
                rh.append_array(cdata[:, h, f], name=f'feature{f}', at='point')
                ru.append_array(cdata[:, h, f], name=f'feature{f}', at='point')
        for f in range(cdata.shape[2]):
            arrName[f, :] = f'feature{f}'

    # extra parameters
    new_qwargs = dict()
    if not 'zoom' in qwargs:
        new_qwargs['zoom'] = 1.7
    if not 'nan_color' in qwargs:
        new_qwargs['nan_color'] = (0, 0, 0, 0)
    new_qwargs.update(qwargs)
    new_size = copy.deepcopy(size)
    new_size[0] = new_size[0] * len(hemis)
    new_size[1] = new_size[1] * cdata.shape[2]
    if 'color_bar' in qwargs:
        new_size[0] = new_size[0] + 60
    p = plot_surf(surfDict, surfList, array_name=arrName, size=new_size, **new_qwargs)
    return p


def surfplot_sub_foldunfold(hippunfold_dir, sub, ses, features, hemis=['L', 'R'], labels=['hipp', 'dentate'],
                            flipRcurv=True, unfoldAPrescale=False, den='0p5mm', modality='T1w', tighten_cwindow=True,
                            rotate=20, resourcesdir=resourcesdir, size=[350, 230], cmap='viridis', **qwargs):
    """
        Plots subject-specific folded and unfolded surfaces (hipp/dentate; folded/unfolded).

        Parameters
        ----------
        hippunfold_dir: str
            Directory path containing unfolded hippocampus data.
        sub: str
            Subject ID. Inputs are path/filenames
        ses: str
            Session ID. Inputs are path/filenames
        features: str
            Feature or measurement to visualize on the surface plot. Can include thickness, curvature, gyrification, subfields,
            or other added data that follows the same naming convention.
        hemis: list of str, optional
            List of hemispheres to visualize. Default is ['L', 'R'].
        labels: list of str, optional
            List of labels for different structures. Default is ['hipp', 'dentate'].
        flipRcurv: bool, optional
            Whether to flip the curvature map for the right hemisphere. Default is True.
        unfoldAPrescale: bool, optional
            Whether to pre-scale the anterior-posterior axis during unfolding. Default is False.
        den: str, optional
            Density parameter for surface plot. Default is '0p5mm'.
        modality: str, optional
            Imaging modality (e.g., 'T1w'). Default is 'T1w'.
        tighten_cwindow: bool, optional
            Whether to tighten the color window for the surface plot. Default is True.
        rotate: bool, optional
            Whether to rotate the surface plot. Default is True.
        resourcesdir: str, optional
            Directory path containing additional resources. Default is the value of resourcesdir.
        size: list of int, optional
            Size of the surface plot. Default is [350, 230].
        cmap: str, optional
            Colormap for the surface plot. Default is 'viridis'.
        **qwargs: dict, optional
            Additional keyword arguments for customization.

        Returns
        -------
        matplotlib.figure.Figure
            The function generates a surface plot for the specified subject's folded and unfolded hippocampus.
        """

    if len(ses) > 0:
        ses = 'ses-' + ses
        uses = '_' + ses
    else:
        uses = ''

    # load surfaces
    surf = []
    n = 0
    for hemi in hemis:
        for space in [modality, 'unfold']:
            for label in labels:
                fn1 = f'{hippunfold_dir}/sub-{sub}/{ses}/surf/sub-{sub}{uses}_hemi-{hemi}_space-{space}_den-{den}_label-{label}_midthickness.surf.gii'
                if glob.glob(fn1):
                    s = read_surface(fn1)
                else:
                    print(fn1 + ' not found, using generic')
                    s = read_surface(
                        f'{resourcesdir}/canonical_surfs/tpl-avg_space-unfold_den-{den}_label-{label}_midthickness.surf.gii')
                    if space != 'unfold': s.Points = np.ones(s.Points.shape) * np.nan
                if label == 'hipp':
                    nptsHipp = s.n_points
                if space == 'unfold':
                    s.Points = s.Points[:, [1, 0, 2]]  # reorient unfold
                    if unfoldAPrescale: s.Points = area_rescale(s.Points, den, label, APaxis=1)
                    if label == 'dentate':
                        s.Points = s.Points + [22, 0, 0]  # translate DG
                if hemi == "L" and space == 'unfold':  # flip L unfolded
                    s.Points[:, 0] = -s.Points[:, 0]
                if label == 'dentate':  # concatenate dentate to hipp (hipp should be immediately prior)
                    s = mc.build_polydata(np.concatenate((oldsurf.Points.copy(), s.Points.copy())),
                                          cells=np.concatenate(
                                              (oldsurf.GetCells2D().copy(), s.GetCells2D().copy() + nptsHipp)))
                oldsurf = s

            surf.append(s)
    npts = len(surf[0].Points)

    # rotate surfaces to approx coronal-oblique
    if rotate:
        #aff = np.loadtxt(f'{resourcesdir}/xfms/corobl-20deg_xfm.txt')
        aff = np.eye(4)
        aff[1,1] = np.cos(np.deg2rad(rotate))
        aff[2,2] = np.cos(np.deg2rad(rotate))
        aff[1,2] = np.sin(np.deg2rad(rotate))
        aff[2,1] = -np.sin(np.deg2rad(rotate))
        for i, h in enumerate(hemis):
            p = np.hstack((surf[i * 2].Points, np.ones((npts, 1))))
            p = p @ aff
            surf[i * 2].Points = p[:, :3]

    # load features data
    ind = [range(nptsHipp), range(nptsHipp, npts)]
    cdata = np.ones([npts, len(hemis), len(features)]) * np.nan
    for h, hemi in enumerate(hemis):
        for l, label in enumerate(labels):
            for f, feature in enumerate(features):
                if feature == 'subfields':
                    type = 'label'
                else:
                    type = 'shape'
                fn2 = f'{hippunfold_dir}/sub-{sub}/{ses}/surf/sub-{sub}{uses}_hemi-{hemi}_space-{modality}_den-{den}_label-{label}_*{feature}*.{type}.gii'
                fn3 = glob.glob(fn2)
                try:
                    cdata[ind[l], h, f] = nib.load(fn3[0]).darrays[0].data
                    if flipRcurv and feature == 'curvature' and hemi == 'L':
                        cdata[ind[l], h, f] = -cdata[ind[l], h, f]
                except:
                    print(fn2 + ' failed')

    # options
    cmaps = np.ones((len(features), len(hemis) * 2), dtype=object)
    cmaps[:] = cmap
    for f, feature in enumerate(features):
        if feature == 'subfields':
            cdata[ind[1], :, f] = np.nanmax(cdata[ind[0], :, f]) + 1
            cmaps[f, :] = ('jet')
        else:
            if tighten_cwindow > 0:
                cdata[:, :, f] = bound_cdata(cdata[:, :, f])

    # set up layout
    surfDict = {}
    surfList = np.ones((len(features), len(hemis) * 2), dtype=object)
    arrName = np.ones((len(features), len(hemis) * 2), dtype=object)
    for h, hemi in enumerate(hemis):
        surfDict.update({f"{hemi}f": surf[h * 2]})
        surfDict.update({f"{hemi}u": surf[h * 2 + 1]})
        if hemi == 'L':
            surfList[:, [h, h + 1]] = np.array([f"{hemi}f", f"{hemi}u"])
        elif hemi == 'R':
            surfList[:, [h * 2, h * 2 + 1]] = np.array([f"{hemi}u", f"{hemi}f"])
        for f, feature in enumerate(features):
            surf[h * 2].append_array(cdata[:, h, f], name=feature, at='point')
            surf[h * 2 + 1].append_array(cdata[:, h, f], name=feature, at='point')
            arrName[f, :] = feature

    # extra parameters
    new_qwargs = dict()
    if not 'zoom' in qwargs:
        new_qwargs['zoom'] = 1.5
    if not 'nan_color' in qwargs:
        new_qwargs['nan_color'] = (0, 0, 0, 0)
    new_qwargs.update(qwargs)
    new_size = copy.deepcopy(size)
    new_size[0] = new_size[0] * len(hemis)
    new_size[1] = new_size[1] * cdata.shape[2]
    if 'color_bar' in qwargs:
        new_size[0] = new_size[0] + 70

    p = plot_surf(surfDict, surfList, array_name=arrName, size=new_size, cmap=cmaps, **new_qwargs)
    return p


def bound_cdata(cdata, cutoff=0.05):
    """
      Returns upper and lower X percent interval values.
      Parameters
      ----------
      cdata :  List of values.
      cutoff : Default is 0.05.
      Returns
      -------
       Values within the upper and lower X percent interval.
    """
    if not cutoff:
        return False
    shp = cdata.shape
    c = cdata.flatten()
    l = np.sort(c[~np.isnan(c)])
    try:
        bounds = l[[int(cutoff * len(l)), int((1 - cutoff) * len(l))]]
        cdata[cdata < bounds[0]] = bounds[0]
        cdata[cdata > bounds[1]] = bounds[1]
    except:
        print('cdata all NaN')
    return np.reshape(cdata, shp)


def area_rescale(vertices, den, label, APaxis=1):
    """
         Rescales unfolded surface vertices to account for overrepresentation of anterior and posterior regions.

         Parameters
         ----------
         vertices : Unfolded surface vertices.
         den : str
             Density of the unfolded space. One of '0p5mm', '1mm', '2mm', or 'unfoldiso'.
         label : str
             Surface label. Either 'hipp' or 'dentate'.
         APaxis : int, optional
             Axis along the anterior-posterior direction. Default is 1.
         Returns
         -------
         numpy.ndarray
             Rescaled unfolded surface vertices.
     """
    w = 126 if label == 'hipp' else 30  # width of unfolded space
    s = 15  # surf area smoothing (sigma)
    Pold = vertices[:, APaxis]

    # compute rescaling from surface areas
    surfarea = \
        nib.load(
            f'{resourcesdir}/canonical_surfs/tpl-avg_space-unfold_den-{den}_label-{label}_surfarea.shape.gii').darrays[
            0].data
    surfarea, _, _ = hippomaps.utils.density_interp(den, 'unfoldiso', surfarea.flatten(), label)
    surfarea = np.reshape(surfarea, (w, 254))
    surfarea = gaussian_filter(surfarea, sigma=s)

    avg_surfarea = np.mean(surfarea, axis=0)
    rescalefactor = np.cumsum(1 / avg_surfarea)
    rescalefactor = rescalefactor - np.min(rescalefactor)
    rescalefactor = rescalefactor / np.max(rescalefactor)
    rescalefactor = rescalefactor + 1 - np.linspace(0, 1, len(rescalefactor))

    rescalefactor = repmat(rescalefactor, w, 1)
    rescalefactor, _, _ = hippomaps.utils.density_interp('unfoldiso', den, rescalefactor.flatten(), label)

    Pnew = Pold * rescalefactor
    vertices[:, APaxis] = Pnew
    return vertices



####################################################################################################

# Translated from matlab: https://github.com/MICA-MNI/micaopen/blob/master/MPC/scripts/build_mpc.m
# Original script by Casey Paquola
# Translated to python by Jessica Royer

# build_mpc   Construct a microstructure profile covariance matrix

# INPUT
# data          surfaces x vertices matrix containing intensity values
# parc          1 x vertices vector with unique integrers correpsonding to
#               node assignment. Leave empty, [], for a vertex-wise MPC
#               Make sure that you provide different parcel # for each hemisphere
#
# OUTPUT
# MPC           microstructural profile covariance matrix
# I             microstructure intensity profiles (node-wise if parc is given)
# problemNodes  which nodes in intensity profiles are made up of NaNs

####################################################################################################

# Import packages
import numpy as np
import scipy.special
import scipy.stats

def build_mpc(data, parc=None, idxExclude=None):
    """
        Construct a microstructure profile covariance matrix.

        Parameters:
        - data (numpy.ndarray): Surfaces x Vertices matrix containing intensity values.
        - parc (numpy.ndarray or None): 1 x Vertices vector with unique integers corresponding to node assignment.
                                       Leave empty (None) for a vertex-wise MPC.
                                       Make sure to provide different parcel numbers for each hemisphere.
        - idxExclude (list or None): Indices of labels to exclude when calculating mean across columns.

        Returns:
        - MPC (numpy.ndarray): Microstructural profile covariance matrix.
        - I (numpy.ndarray): Microstructure intensity profiles (node-wise if parc is given).
        - problemNodes (list or int): List of indices where intensity profiles have NaN values or 0 if no issue.

        Notes:
        If parc is not provided, MPC will be computed vertex-wise.
        The function handles parcellation, averages profiles within nodes, and detects outliers.

        Example:
        ```python
        data = np.random.rand(10, 100)  # Example data
        parc = np.random.randint(1, 5, 100)  # Example parcel information
        MPC, I, problemNodes = build_mpc(data, parc)
        ```

        Original script by Casey Paquola, translated to Python by Jessica Royer.
        Translated from MATLAB: https://github.com/MICA-MNI/micaopen/blob/master/MPC/scripts/build_mpc.m
        """
    # If no parcellation is provided, MPC will be computed vertexwise
    if parc is None:
        downsample = 0
    else:
        downsample = 1


    # Parcellate input data according to parcellation scheme provided by user
    if downsample == 1:
        uparcel = np.unique(parc)
        I = np.zeros([data.shape[0], len(uparcel)])

        # Parcellate data by averaging profiles within nodes
        for (ii, _) in enumerate(uparcel):

            # Get vertices within parcel
            thisparcel = uparcel[ii]
            tmpData = data[:, parc == thisparcel]
            tmpData[:,np.mean(tmpData) == 0] = 0

            # Define function to find outliers: Return index of values above three scaled median absolute deviations of input data
            # https://www.mathworks.com/help/matlab/ref/isoutlier.html
            def find_outliers(data_vector):
                c = -1 / (np.sqrt(2) * scipy.special.erfcinv(3/2))
                scaled_MAD = c * np.median(np.abs(data_vector - np.median(data_vector)))
                is_outlier = np.greater(data_vector, (3 * scaled_MAD) + np.median(data_vector))
                idx_outlier = [i for i, x in enumerate(is_outlier) if x]
                return idx_outlier

            # Find if there are any outliers in vertex-wise average profile within given parcel
            idx = find_outliers(np.mean(tmpData, axis = 0))
            if len(idx) > 0:
                tmpData[:,idx] = np.nan

            # Average profiles within parcels
            I[:,ii] = np.nanmean(tmpData, axis = 1)

        # Get matrix sizes
        szI = I.shape
        szZ = [len(uparcel), len(uparcel)]

    else:
        I = data
        szI = data.shape
        szZ = (data.shape[1], data.shape[1])


    # Build MPC
    if np.isnan(np.sum(I)):

        # Find where are the NaNs
        is_nan = np.isnan(I[1,:])
        problemNodes = [i for i, x in enumerate(is_nan) if x]
        print("")
        print("---------------------------------------------------------------------------------")
        print("There seems to be an issue with the input data or parcellation. MPC will be NaNs!")
        print("---------------------------------------------------------------------------------")
        print("")

        # Fill matrices with NaN for return
        I = np.zeros(szI)
        I[I == 0] = np.nan
        MPC = np.zeros(szZ)
        MPC[MPC == 0] = np.nan

    else:
        problemNodes = 0

        # Calculate mean across columns, excluding mask and any excluded labels input
        I_mask = I
        NoneType = type(None)
        if type(idxExclude) != NoneType:
            for i in idxExclude:
                I_mask[:, i] = np.nan
        I_M = np.nanmean(I_mask, axis = 1)

        # Get residuals of all columns (controlling for mean)
        I_resid = np.zeros(I.shape)
        for c in range(I.shape[1]):
            y = I[:,c]
            x = I_M
            slope, intercept, _, _, _ = scipy.stats.linregress(x,y)
            y_pred = intercept + slope*x
            I_resid[:,c] = y - y_pred

        R = np.corrcoef(I_resid, rowvar=False)

        # Log transform
        MPC = 0.5 * np.log( np.divide(1 + R, 1 - R) )
        MPC[np.isnan(MPC)] = 0
        MPC[np.isinf(MPC)] = 0

        # CLEANUP: correct diagonal and round values to reduce file size
        # Replace all values in diagonal by zeros to account for floating point error
        for i in range(0,MPC.shape[0]):
                MPC[i,i] = 0
        # Replace lower triangle by zeros
        MPC = np.triu(MPC)

    # Output MPC, microstructural profiles, and problem nodes
    return (MPC, I, problemNodes)



import numpy as np
import glob
import os
import sys
import copy
from joblib import Parallel, delayed
import warnings
from scipy.interpolate import griddata
from scipy.interpolate import NearestNDInterpolator
import nibabel as nib
from scipy.interpolate import interp1d
from scipy.interpolate import interpn
from numpy.matlib import repmat
import pygeodesic.geodesic as geodesic
from pathlib import Path
resourcesdir = str(Path(__file__).parents[1]) + '/resources'

def blockPrint():
    sys.stdout = open(os.devnull, 'w')

def enablePrint():
    sys.stdout = sys.__stdout__

def avg_neighbours(F, cdat, n):
    """
    Averages vertex-wise data at vertex n with its neighboring vertices.

    Parameters
    ----------
    F: Faces, numpy.ndarray
    cdat: vertex data, numpy.ndarray
    n: Vertex index for which the data will be averaged with its neighboring vertices, int
    Returns
    -------
    float
        The average value of the vertex-wise data at vertex n and its neighboring vertices.
    """
    frows = np.where(F == n)[0]
    v = np.unique(F[frows, :])
    cdat = np.reshape(cdat, (len(cdat), -1))
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        out = np.nanmean(cdat[v,:], 0)
    return out


def surfdat_smooth(F, cdata, iters=1, cores=1):
    """
       Smooths surface data across neighboring vertices.
       TODO: convert to mm vis calibration curves in https://github.com/MELDProject/meld_classifier/blob/9d3d364de86dc207d3a1e5ec11dcab3ef012ebcb/meld_classifier/mesh_tools.py#L17'''
       This function assumes that vertices are evenly spaced and evenly connected.

       Parameters
       ----------
       F : numpy.ndarray
       cdata : numpy.ndarray
       iters : int, optional
       cores : int, optional
       Returns
       -------
       numpy.ndarray
           Smoothed surface data across neighboring vertices.
       """
    sz = cdata.shape
    cdata = cdata.reshape(cdata.shape[0], -1)
    cdata_smooth = copy.deepcopy(cdata)
    for i in range(iters):
        if cores>1:
            c = Parallel(n_jobs=cores, backend='multiprocessing')(delayed(avg_neighbours)(F, cdata, n) for n in range(len(cdata)))
            cdata_smooth = np.array(c)
        else:
            for n in range(len(cdata)):
                cdata_smooth[n,:] = avg_neighbours(F, cdata, n)
        cdata = copy.deepcopy(cdata_smooth)
    return cdata_smooth.reshape(sz)


def profile_align(P, patchdist=None, V=[],F=[], maxroll=5):
    """
       Aligns microstructural profiles in the depth direction across a set of surfaces.

       Parameters
       ----------
       P : A VxD matrix of intensities (vertices x depths).
       V : The midthickness surface vertices.
       F : The midthickness surface faces.
       patchdist : Radius (in mm) of geodesic distance to compute the average profile.
           If None, then all profiles are used. Default is None.
       maxroll : int
           Maximum shift.
       Returns
       -------
       A matrix the same size as P, pads profiles by maxroll, then rolls them by +/- maxroll until maximum overlap
           with the patch average is achieved.

       Notes
       -----
       - If patchdist is None, the function aligns profiles based on the mean profile across all vertices.
       - If patchdist is specified, the function aligns profiles based on the mean profile within a geodesic patch around each vertex.
       - The alignment is performed by rolling the profiles to achieve maximum correlation with the reference profile.
       """
    P = np.pad(P, ((0, 0), (maxroll, maxroll)), mode='edge')
    Paligned = np.ones(P.shape) * np.nan

    if patchdist == None:
        p_mean = np.nanmean(P, axis=0)
        # get R for all rolls
        rolls = np.arange(-maxroll, maxroll + 1)
        R = np.zeros((len(P), len(rolls)))
        for o, offset in enumerate(rolls):
            p_mean_off = np.reshape(np.roll(p_mean, offset), (1, len(p_mean)))
            R[:, o] = np.corrcoef(p_mean_off, P)[1:, 0]
        Rbest = np.argmax(R, axis=1)
        # keep only the best roll
        for v in range(P.shape[0]):
            Paligned[v, :] = np.roll(P[v, :], -rolls[Rbest[v]])
    else:
        geoalg = geodesic.PyGeodesicAlgorithmExact(V, F)
        for v in range(len(V)):
            # use a patch around the given vertex as the averaged reference
            dist, _ = geoalg.geodesicDistances(np.array([v]), None)
            p_mean = np.nanmean(P[dist < patchdist, :], axis=0)
            rolls = np.arange(-maxroll, maxroll + 1)
            R = np.zeros((len(rolls)))
            for o, offset in enumerate(rolls):
                p_mean_off = np.reshape(np.roll(p_mean, offset), (1, len(p_mean)))
                R[o] = np.corrcoef(p_mean_off, P[v, :])[0, 1]
            Rbest = np.argmax(R)
            Paligned[v, :] = np.roll(P[v, :], -rolls[Rbest])
    return Paligned[:, maxroll:-maxroll]


def laplace_solver(faces, init, maxiters=1e4, conv=1e-6, cores=8):
    """
        Solves the Laplace equation along vertices of a surface.

        Parameters
        ----------
        faces : Faces of a surface mesh.
        init : Initial vertex-wise values. Source(s) should have a value of 0, Sink(s) should have a value of 1. NaN values will be masked. Time required scales with the number of unmasked vertices.
        maxiters : int, optional
            Maximum iterations to run. Default is 1e4.
        conv : float, optional
            Convergence criterion. Default is 1e-6.
        cores : int, optional
            Number of CPU cores to use (per iteration). Default is 8.
        Returns
        -------
        LP : Laplace solution (vertex-wise).
        change : sum of absolute changes per iteration (should converge towards 0)
        """
    ind_start = np.where(init == 0)[0]
    ind_end = np.where(init == 1)[0]
    mask = np.where(np.isnan(init))[0]
    # optimize speed
    faces = np.delete(faces, np.where(np.isin(faces, mask))[0], 0)
    change = []
    LP = copy.deepcopy(init)
    for iters in range(maxiters):
        LPup = surfdat_smooth(faces, LP, iters=1, cores=cores)
        LPup[ind_start] = 0
        LPup[ind_end] = 1
        LPup[mask] = np.nan
        c = np.nansum(np.abs(LP - LPup))
        change.append(c)
        LP = copy.deepcopy(LPup)
        if c < conv:
            break
    return LP, change


def fillnanvertices(F, V):
    """
       Fills NaNs by iteratively computing the nanmean of nearest neighbors until no NaNs remain.
       Can be used to fill missing vertices OR missing vertex cdata.

       Parameters
       ----------
       F : Faces of a surface mesh.
       V : Vertex coordinates or vertex-wise data with potential NaN values.

       Returns
       -------
       The array with NaN values filled by the nanmean of nearest neighbors.
       """
    Vold = copy.deepcopy(V)
    Vnew = copy.deepcopy(V)
    while np.isnan(np.sum(Vnew)):
        # index of vertices containing nan
        vrows = np.unique(np.where(np.isnan(Vnew))[0])
        # replace with the nanmean of neighbouring vertices
        for n in vrows:
            frows = np.where(F == n)[0]
            neighbours = np.unique(F[frows, :])
            Vnew[n] = np.nanmean(Vnew[neighbours], 0)
        if sum(np.isnan(Vold)) == sum(np.isnan(Vnew)):  # stop if no changes
            break
        else:
            Vold = Vnew
    return Vnew


def density_interp(indensity, outdensity, cdata, label, method='linear', resourcesdir=resourcesdir):
    """
       Interpolates data from one surface density onto another via unfolded space.

       Parameters
       ----------
       indensity : str
           One of '0p5mm', '1mm', '2mm', or 'unfoldiso'.
       outdensity : str
           One of '0p5mm', '1mm', '2mm', or 'unfoldiso'.
       cdata : numpy.ndarray (true)
           Data to be interpolated (same number of vertices, N, as indensity).
       label : str
           'hipp' or 'dentate'.
       method : str, optional
           Interpolation method. Options are 'nearest', 'linear', or 'cubic'. Default is 'linear'. (true?)
       resourcesdir : str, optional
           Path to the hippunfold resources folder.

       Returns
       -------
       interp : interpolated data
       faces: face connectivity from new surface density
    """
    VALID_STATUS = {'0p5mm', '1mm', '2mm', 'unfoldiso'}
    if indensity not in VALID_STATUS:
        raise ValueError("results: indensity must be one of %r." % VALID_STATUS)
    if outdensity not in VALID_STATUS:
        raise ValueError("results: outdensity must be one of %r." % VALID_STATUS)

    # load unfolded surfaces for topological matching
    startsurf = nib.load(
        f'{resourcesdir}/canonical_surfs/tpl-avg_space-unfold_den-{indensity}_label-{label}_midthickness.surf.gii')
    vertices_start = startsurf.get_arrays_from_intent('NIFTI_INTENT_POINTSET')[0].data
    targetsurf = nib.load(
        f'{resourcesdir}/canonical_surfs/tpl-avg_space-unfold_den-{outdensity}_label-{label}_midthickness.surf.gii')
    vertices_target = targetsurf.get_arrays_from_intent('NIFTI_INTENT_POINTSET')[0].data
    faces = targetsurf.get_arrays_from_intent('NIFTI_INTENT_TRIANGLE')[0].data

    # interpolate
    interp = griddata(vertices_start[:, :2], values=cdata, xi=vertices_target[:, :2], method=method)
    # fill any NaNs
    interp = fillnanvertices(faces, interp)
    return interp, faces, vertices_target


def surface_to_volume(surf_data, indensity, hippunfold_dir, sub, ses, hemi, space='*', label='hipp', save_out_name=None,
                      method='nearest'):
    """
       Labels voxels using data on a folded/unfolded surface and native space coordinates images.
        from https://github.com/khanlab/hippunfold/blob/master/hippunfold/workflow/scripts/label_subfields_from_vol_coords.py
        this function labels voxels using data on a folded/unfolded surface (midthickness or any), and native space coords (ap, pd) images
        TODO: consider trying to simplify inputs specifying the coords paths?

       Parameters
       ----------
       surf_data : Data on the folded/unfolded surface (midthickness or any).
       indensity : str
           Density of the unfolded space. One of '0p5mm', '1mm', '2mm', or 'unfoldiso'.
       hippunfold_dir : str
           Directory path to the HippUnfold output.
       sub : str
           Subject identifier.
       ses : str
           Session identifier.
       hemi : str
           Hemisphere. Either 'L' or 'R'.
       space : str, optional
            Default is '*'.
       label : str, optional
            Default is 'hipp'.
       save_out_name : str, optional
           Output file name to save the label image. Default is None.
       method : str, optional
           Interpolation method. Options are 'nearest', 'linear', or 'cubic'. Default is 'nearest'.
       Returns
       -------
       numpy.ndarray
           Labeled voxel data.
       """
    if len(ses) > 0:
        ses = 'ses-' + ses
        uses = '_' + ses
    else:
        uses = ''

    nii_ap = glob.glob(
        f'{hippunfold_dir}/sub-{sub}/{ses}/coords/sub-{sub}{uses}_dir-AP_hemi-{hemi}_space-{space}_label-{label}_desc-laplace_coords.nii.gz')[
        0]
    nii_pd = glob.glob(
        f'{hippunfold_dir}/sub-{sub}/{ses}/coords/sub-{sub}{uses}_dir-PD_hemi-{hemi}_space-{space}_label-{label}_desc-laplace_coords.nii.gz')[
        0]
    nii_mask = \
        glob.glob(f'{hippunfold_dir}/sub-{sub}/{ses}/anat/sub-{sub}{uses}_hemi-{hemi}_space-{space}_*_dseg.nii.gz')[0]

    # resample surface data into APxPD shape
    surf_data[np.isnan(surf_data)] = -999
    if indensity != 'unfildiso':
        surf_data_unfoldiso, _, _ = density_interp(indensity, 'unfoldiso', surf_data, label, method=method)
    surf_data_unfoldiso = surf_data_unfoldiso.reshape(126, 254).T
    surf_data_unfoldiso[surf_data_unfoldiso == -999] = np.nan

    # undo unfolded space warp
    warp = glob.glob(
        f'{hippunfold_dir}/../work/sub-{sub}/{ses}/warps/sub-{sub}{uses}_hemi-{hemi}_space-unfold_desc-SyN_from-*_to-subject_type-itk_xfm.nii.gz')
    if not warp: raise Warning(
        "No unfolded space warp found. It may be that your HippUnfold output work dir is tarred, or you ran hippunfold without unfolded space registration. Proceeding without unfolded space registration")
    try:
        t = nib.load(warp[0])
        tmp_nib = nib.Nifti1Image(surf_data_unfoldiso, t.affine)
        nib.save(tmp_nib, "tmp.nii.gz")
        if method == "nearest":
            meth = "NearestNeighbor"
        elif method == "linear":
            meth = "Linear"
        t = os.system(f"antsApplyTransforms -i tmp.nii.gz -r tmp.nii.gz -o tmpWarped.nii.gz -t {warp[0]} -n {meth}")
        if t != 0: raise Exception("ANTs not found or returning error")
        surf_data_unfoldiso = nib.load("tmpWarped.nii.gz").get_fdata()[:, :, 0]
        os.system("rm tmp.nii.gz tmpWarped.nii.gz")
    except:
        raise Warning("error in unfolded space registration")

    # setup the interpolating grid
    spacing_ap = np.linspace(0, 1, 254)
    spacing_pd = np.linspace(0, 1, 126)
    points = (spacing_ap, spacing_pd)

    # load up the coords
    ap_nib = nib.load(nii_ap)
    pd_nib = nib.load(nii_pd)
    ap_img = ap_nib.get_fdata()
    pd_img = pd_nib.get_fdata()

    # get mask of coords
    mask_nib = nib.load(nii_mask)
    mask_img = mask_nib.get_fdata()
    if label == 'hipp':
        mask = np.logical_and(mask_img > 0, mask_img < 6)
    elif label == 'dentate':
        mask = mask_img == 6
    else:
        mask = mask_img > 0

    # interpolate
    query_points = np.vstack((ap_img[mask], pd_img[mask])).T
    labelled_points = interpn(points, surf_data_unfoldiso, query_points, method=method)

    # put back into image
    label_img = np.zeros(ap_img.shape, dtype="uint16")
    label_img[mask == 1] = labelled_points

    if save_out_name:
        # save label img
        label_nib = nib.Nifti1Image(label_img, ap_nib.affine, ap_nib.header)
        nib.save(label_nib, save_out_name)
    return label_img




#!/usr/bin/env python

"""A setuptools based setup module.
See:
https://packaging.python.org/guides/distributing-packages-using-setuptools/
"""

import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="HippoMaps",
    version="0.1.5",
    author="Jordan DeKraker",
    author_email="jordandekraker@gmail.com",
    description="A toolbox for viewing, manipulating, and additional actions on HippUnfold outputs",
    long_description=long_description,
    long_description_content_type="text/x-rst",
    url="https://github.com/jordandekraker/hippomaps",
    packages=setuptools.find_packages(),
    license="BSD 3-Clause License",
    package_data={
        "": ["resources/*"],
    },
    include_package_data=True,
    classifiers=[
        "Intended Audience :: Science/Research",
        "Intended Audience :: Developers",
        "Topic :: Software Development :: Build Tools",
        "Topic :: Scientific/Engineering",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
    ],
    keywords='hippunfold registration',
    python_requires=">=3.7.0",
    install_requires=[
        "Path>=16.4.0",
        "joblib",
        "brainspace>=0.1.2",
        "nibabel>=3.2.2",
        "nilearn>=0.7.0",
        "numpy>=1.23.1,<2.0.0",
        "pandas>=0.23",
        "scipy>=1.3.3",
        "matplotlib>=2.0.0",
        "pygeodesic>=0.1.8",
        "wget>=3.2",
        "eigenstrapping>=0.0.1.10",
	"pathlib",
	"parspin",
	"tabulate",
	"umap-learn>=0.5.5",
    ],
    extras_require={"dev": ["gitpython", "hcp-utils", "mypy", "plotly", "pytest"]},
    zip_safe=False,
)



import nibabel as nib
import urllib.request, json 
import wget
from fnmatch import fnmatch
from pathlib import Path
from collections import deque
import os
resourcesdir = str(Path(__file__).parents[1]) + '/resources/'
downloads = resourcesdir + '/downloads/'
checkpointsdir = str(Path(__file__).parents[1]) + '/tutorials/checkpoints/'

repoID = '92p34'

# first get all available files (and their hash) from OSF

name_hashes={}
maxdepth=4
current_hash=''
def tree_osf_hashes(repoID, name_hashes, maxdepth, current_hash):
    """
    Iteratively fetch file hashes from the OSF repository using a queue to manage folders. Each folder's pages
    are handled within its context in the queue to ensure complete coverage of files.

    Parameters
    ----------
    repoID : str
        The ID of the repository.
    name_hashes : dict
        A dictionary to store file names and their corresponding hashes.
    maxdepth : int
        The maximum depth to traverse the file tree.

    Returns
    -------
    None
        Populates the name_hashes dictionary with file names and their corresponding hashes.
    """
    queue = deque([(0, '')])  # Initialize the queue with the starting point: depth and root hash

    while queue:
        depth, folder_hash = queue.popleft()
        if depth > maxdepth:
            continue  # Skip deeper levels than the max allowed

        next_urls = deque([f"https://api.osf.io/v2/nodes/{repoID}/files/osfstorage/{folder_hash}"])
        try:
            while next_urls:  # This loop is to handle pagination within the current folder
                next_url = next_urls.popleft()
                with urllib.request.urlopen(next_url) as url:
                    response = json.load(url)
                    data = response['data']

                    for item in data:
                        materialized_path = item['attributes']['materialized_path']
                        item_hash = item['attributes']['path']
                        kind = item['attributes']['kind']
                        name_hashes[materialized_path] = item_hash
                        
                        # Enqueue subfolders for further exploration, but only if within depth limits
                        if kind == 'folder' and depth + 1 <= maxdepth:
                            queue.append((depth + 1, item_hash))

                    # Safely fetch the next page if available
                    next_url = response.get('links', {}).get('next', {}) if isinstance(response.get('links'), dict) else {}
                    if next_url:
                        next_urls.append(next_url)

        except Exception as e:
            print(f"Failed to fetch or parse data for URL {next_url}: {str(e)}")
            continue  # Continue with the next item in the queue if an error occurs

tree_osf_hashes(repoID,name_hashes,maxdepth,current_hash)

# utility for making dict searchable
class GlobDict(dict):
    def glob(self, match):
        """
       Glob-style pattern matching for the dictionary keys.@match should be a glob style pattern match (e.g. '*.txt')
       Parameters
       ----------
       match : str
           The glob-style pattern to match keys.
       Returns
       -------
       dict
           A filtered dictionary with keys matching the specified pattern.
        """
        return dict([(k,v) for k,v  in self.items() if fnmatch(k, match)])
glob_dict = GlobDict( **name_hashes )


# now the main function of interest
def get_map(repo='HippoMaps-initializationMaps',Dataset='*',Method='*',Modality='*', hemi='*', den='*', extension='.shape.gii'):
    """
    Search and load data from OSF with specified parameters.
    See https://osf.io/92p34/ for examples

    Parameters
    ----------
    repo : str, optional
        Top-level repository name (default: 'HippoMaps-initializationMaps').
    Dataset : str, optional
        Source of the map (default: '*').
    Method : str, optional
        Method being used (e.g., 'MRI', 'MRI-7T', or 'histology') (default: '*').
    Modality : str, optional
        Modality within the Method (e.g., 'qT1', 'FA', or 'Merker') (default: '*').
    hemi : str, optional
        Hemisphere (e.g., 'L', '*', or 'mix') (default: '*').
    den : str, optional
        Only search for data that is already at a given density (e.g., '0p5mm' or 'unfoldiso'). Note that data can always be resampled to a different density after loading (default: '*').
    extension : str, optional
        File extension (e.g., '.shape.gii', '.func.gii', or '.label.gii') (default: '.shape.gii').

    Returns
    -------
    data : list of arrays
        List of arrays containing data that matched search terms.
    names : list of str
        List of filenames that matched search terms.
    """
    searchstr = f'/{repo}/Dataset-{Dataset}/{Method}-{Modality}*hemi-{hemi}*den-{den}*{extension}'
    globbed_results = glob_dict.glob(searchstr)
    data = []
    names = [*globbed_results]
    for fn, h in globbed_results.items():
        if not os.path.isdir(downloads + os.path.dirname(fn)):
            os.makedirs(downloads + os.path.dirname(fn))
        if not os.path.isfile(downloads + fn):
            wget.download(f"https://files.ca-1.osf.io/v1/resources/{repoID}/providers/osfstorage{h}",out=downloads + fn)
        data.append(nib.load(downloads + fn).darrays[0].data)
    return data, names


def get_tutorialCheckpoints(filenames):
    """
    Downloads checkpoint data needed for running hippomaps tutorials. 
    filenames : list of strings
        Filenames to be downloaded
    """
    if not os.path.isdir(checkpointsdir):
        os.mkdir(checkpointsdir)
    for fn in filenames:
        if not os.path.isfile(checkpointsdir + fn):
            h = name_hashes['/tutorialCheckpoints/' + fn]
            wget.download(f"https://files.ca-1.osf.io/v1/resources/{repoID}/providers/osfstorage{h}",out=checkpointsdir + fn)